import builtins as __builtin__

import socket
import time
import inspect
from maggy import util, tensorboard, constants
from maggy.core import rpc, exceptions, config
from maggy.core.reporter import Reporter
from pyspark import TaskContext

from hops import hdfs as hopshdfs
import tensorflow as tf

if config.tf_version >= 2:
    from tensorboard.plugins.hparams import api_pb2
    from tensorboard.plugins.hparams import summary
    from tensorboard.plugins.hparams import summary_v2


def _prepare_func(app_id, run_id, experiment_type, map_fun, server_addr, hb_interval, secret, app_dir):

    def _wrapper_fun(iter):
        """
        Wraps the user supplied training function in order to be passed to the
        Spark Executors.

        Args:
            iter:

        Returns:

        """

        # get task context information to determine executor identifier
        partition_id, task_attempt = util.get_partition_attempt_id()

        client = rpc.Client(server_addr, partition_id,
                            task_attempt, hb_interval, secret)
        log_file = app_dir + '/logs/executor_' + str(partition_id) + '_' + str(task_attempt) + '.log'

        # save the builtin print
        original_print = __builtin__.print

        reporter = Reporter(log_file, partition_id, task_attempt, original_print)

        def maggy_print(*args, **kwargs):
            """Maggy custom print() function."""
            reporter.log(' '.join(str(x) for x in args))
            original_print(*args, **kwargs)

        # override the builtin print
        __builtin__.print = maggy_print

        try:
            client_addr = client.client_addr

            host_port = client_addr[0] + ":" + str(client_addr[1])

            exec_spec = {}
            exec_spec['partition_id'] = partition_id
            exec_spec['task_attempt'] = task_attempt
            exec_spec['host_port'] = host_port
            exec_spec['trial_id'] = None

            reporter.log("Registering with experiment driver", False)
            client.register(exec_spec)

            # blocking
            # _ = client.await_reservations()

            client.start_heartbeat(reporter)

            # blocking
            # XXX separate suggestion calls for different types?
            trial_id, parameters = client.get_suggestion()

            while not client.done:
                if experiment_type == 'ablation':
                    parameters.pop('ablated_feature')
                    parameters.pop('ablated_layer')

                reporter.set_trial_id(trial_id)

                tb_logdir = app_dir + '/trials/' + trial_id
                tensorboard._register(tb_logdir)
                hopshdfs.mkdir(tb_logdir)

                try:
                    reporter.log("Starting Trial: {}".format(trial_id), False)
                    reporter.log("Trial Configuration: {}".format(parameters), False)

                    sig = inspect.signature(map_fun)
                    if sig.parameters.get('reporter', None):
                        retval = map_fun(**parameters, reporter=reporter)
                    else:
                        retval = map_fun(**parameters)

                    # Make sure user function returns a numeric value
                    if retval is None:
                        reporter.log(
                            "ERROR: Training function can't return None", False)
                        raise Exception("Training function can't return None")
                    elif not isinstance(retval, constants.USER_FCT.RETURN_TYPES):
                        reporter.log(
                            "ERROR: Training function returns non numeric value: {}"
                            .format(type(retval)), False)
                        raise Exception(
                            "ERROR: Training function returns non numeric value: {}"
                            .format(type(retval)))

                except exceptions.EarlyStopException as e:
                    retval = e.metric
                    reporter.log("Early Stopped Trial.", False)
                finally:
                    client.finalize_metric(retval, reporter)
                    reporter.log("Finished Trial: {}".format(trial_id), False)
                    reporter.log("Final Metric: {}".format(retval), False)

                # blocking
                trial_id, parameters = client.get_suggestion()

        except:
            reporter.fd.close()
            raise
        finally:
            reporter.fd.close()
            client.stop()
            client.close()

    return _wrapper_fun



class AblationStudy(object):
    """The `AblationStudy` object is the entry point to define an ablation
    study with maggy. This object can subsequently be passed as an argument
    when the experiment is launched with `experiment.lagom()`.

    Sample usage:

    >>> from maggy.ablation import AblationStudy
    >>> ablation_study = AblationStudy('titanic_train_dataset',
    >>>     label_name='survived')

    Define your study by including layers and features, which should be
    ablated:

    >>> ablation_study.features.include('pclass', 'fare')
    >>> ablation_study.model.layers.include('my_dense_two',
    >>>     'my_dense_three')

    You can also add a layer group using a list:

    >>> ablation_study.model.layers.include_groups(['my_dense_two',
    >>>     'my_dense_four'])

    Or add a layer group using a prefix:

    >>> ablation_study.model.layers.include_groups(prefix='my_dense')

    Next you should define a base model function using the layer and feature
    names you previously specified:

    >>> # you only need to add the `name` parameter to layer initializers
    >>> def base_model_generator():
    >>>     model = tf.keras.Sequential()
    >>>     model.add(tf.keras.layers.Dense(64, activation='relu'))
    >>>     model.add(tf.keras.layers.Dense(..., name='my_dense_two', ...)
    >>>     model.add(tf.keras.layers.Dense(32, activation='relu'))
    >>>     model.add(tf.keras.layers.Dense(..., name='my_dense_sigmoid', ...)
    >>>     # output layer
    >>>     model.add(tf.keras.layers.Dense(1, activation='linear'))
    >>>     return model

    Make sure to include the generator function in the study:

    >>> ablation_study.model.set_base_model_generator(base_model_generator)

    Last but not least you can define your actual training function:

    >>> from maggy import experiment
    >>> from maggy.callbacks import KerasBatchEnd
    ​
    >>> def training_function(dataset_function, model_function, reporter):
    >>>     import tensorflow as tf
    >>>     epochs = 5
    >>>     batch_size = 10
    >>>     tf_dataset = dataset_function(epochs, batch_size)
    >>>     model = model_function()
    >>>     model.compile(optimizer=tf.train.AdamOptimizer(0.001),
    >>>             loss='binary_crossentropy',
    >>>             metrics=['accuracy'])
    >>>     ### Maggy REPORTER
    >>>     callbacks = [KerasBatchEnd(reporter, metric='acc')]
    >>>     history = model.fit(tf_dataset, epochs=5, steps_per_epoch=30)
    >>>     return float(history.history['acc'][-1])

    Lagom the experiment:

    >>> result = experiment.lagom(map_fun=training_function,
    >>>                         experiment_type='ablation',
    >>>                         ablation_study=ablation_study,
    >>>                         ablator='loco',
    >>>                         name='Titanic-LOCO',
    >>>                         hb_interval=5)
    """

    def __init__(
        self, training_dataset_name, training_dataset_version, label_name,
        **kwargs):
        """Initializes the ablation study.

        :param training_dataset_name: Name of the training dataset in the
            featurestore.
        :type training_dataset_name: str
        :param training_dataset_version: Version of the training dataset to be
            used.
        :type training_dataset_version: int
        :param label_name: Name of the target prediction label.
        :type label_name: str
        """
        self.features = Features()
        self.model = Model()
        self.hops_training_dataset_name = training_dataset_name
        self.hops_training_dataset_version = training_dataset_version
        self.label_name = label_name
        self.custom_dataset_generator = kwargs.get('dataset_generator', False)

    def to_dict(self):
        """
        Returns the ablation study configuration as a Python dictionary.

        :return: A dictionary with ablation study configuration parameters as
            keys (i.e. 'training_dataset_name', 'included_features', etc.)
        :rtype: dict
        """
        ablation_dict = {
            'training_dataset_name': self.hops_training_dataset_name,
            'training_dataset_version': self.hops_training_dataset_version,
            'label_name': self.label_name,
            'included_features': list(self.features.included_features),
            'included_layers': list(self.model.layers.included_layers),
            'custom_dataset_generator': True if self.custom_dataset_generator else False,
        }

        return ablation_dict


class Features(object):
    def __init__(self):
        self.included_features = set()

    def include(self, *args):
        """
        Add one or several features of the dataset to the ablation study. Features can be included one by one or
        as a list of strings.
        :param args: Strings or lists of strings, that should match feature names.
        :type args: str or list
        :return:
        """
        for arg in args:
            if type(arg) is list:
                for feature in arg:
                    self._include_single_feature(feature)
            else:
                self._include_single_feature(arg)

    def _include_single_feature(self, feature):
        if type(feature) is str:
            self.included_features.add(feature)
        else:
            raise ValueError(
                "features.include() only accepts strings or lists of strings, "
                "but it received {0} which is of type '{1}'."
                .format(str(feature), type(feature).__name__))

    def exclude(self, *args):
        """
        Exclude one or several features of the dataset from the list of features that have been included in the
        ablation study. Features can be excluded one by one or as a list of strings. This method will check to see
        if the feature has already been included in the study before excluding it.
        :param args: Strings or lists of strings, that should match feature names.
        :type args: str or list
        :return:
        """
        for arg in args:
            if type(arg) is list:
                for feature in arg:
                    self._exclude_single_feature(feature)
            else:
                self._exclude_single_feature(arg)

    def _exclude_single_feature(self, feature):
        if type(feature) is str:
            if feature in self.included_features:
                self.included_features.remove(feature)
                print("Feature '{0}' is excluded from the ablation study.".format(str(feature)))
        else:
            raise ValueError("features.exclude() only accepts strings or lists of strings, "
                             "but it received {0} (of type '{1}')."
                             .format(str(feature), type(feature).__name__))

    def list_all(self):
        for feature in self.included_features:
            print(feature)


class Model(object):
    def __init__(self):
        self.layers = Layers()
        self.base_model_generator = None

    def set_base_model_generator(self, base_model_generator):
        self.base_model_generator = base_model_generator


class Layers(object):
    def __init__(self):
        self.included_layers = set()
        self.included_groups = set()

    def include(self, *args):
        """
        Include layers in the ablations study. Note that the first (input) and the last (output) layer of the base model
        can never be included in the ablation study.
        :param args: Strings or lists of strings, that should match layer names.
        :type args: str or list
        :return:
        """
        for arg in args:
            if type(arg) is list:
                for layer in arg:
                    self._include_single_layer(layer)
            else:
                self._include_single_layer(arg)

    def _include_single_layer(self, layer):
        if type(layer) is str:
            self.included_layers.add(layer)
            # print("included {}".format(layer))  # this still prints even if was duplicate
        else:
            raise ValueError("layers.include() only accepts strings or lists of strings, "
                             "but it received {0} which is of type '{1}'."
                             .format(str(layer), type(layer).__name__))

    def exclude(self, *args):
        for arg in args:
            if type(arg) is list:
                for layer in arg:
                    self._exclude_single_layer(layer)
            else:
                self._exclude_single_layer(arg)

    def _exclude_single_layer(self, layer):
        if type(layer) is str:
            if layer in self.included_layers:
                self.included_layers.remove(layer)
        else:
            raise ValueError("layers.exclude() only accepts strings or lists of strings, "
                             "but it received {0} (of type '{1}')."
                             .format(str(layer), type(layer).__name__))

    def include_groups(self, *args, prefix=None):
        """
        Adds a group of layers that should be removed from the model together. The groups are specified either
        by being passed as a list of layer names (strings), or a string as a common prefix of their layer names.
        Each list of strings would result in a single grouping.
        :param prefix: A string that is a prefix of the names of a group of layers in the base model.
        :type prefix: str
        :param args: Lists of strings (layer names) to indicate groups of layers. The length of the list should be
        greater than one - it does not make sense to have a group of layers that consists of only one layer.
        :type args: list
        """
        if prefix is not None:
            if type(prefix) is str:
                self.included_groups.add(frozenset([prefix]))
            else:
                raise ValueError("`prefix` argument of layers.include_groups() should either be "
                                 "a `NoneType` or a `str`, but it received {0} (of type '{1}'."
                                 .format(str(prefix), type(prefix).__name__))

        for arg in args:
            if type(arg) is list and len(arg) > 1:
                self.included_groups.add(frozenset(arg))
            elif type(arg) is list and len(arg) == 1:
                raise ValueError("layers.include_groups() received a list ( {0} ) "
                                 "with only one element: Did you want to include a single layer in "
                                 "your ablation study? then you should use layers.include()."
                                 .format(str(arg)))
            else:
                raise ValueError("layers.include_groups() only accepts a prefix string, "
                                 "or lists (with more than one element) "
                                 "of strings, but it received {0} (of type '{1}')."
                                 .format(str(arg), type(arg).__name__))

    def exclude_groups(self, *args, prefix=None):
        """
        Removes a group of layers from being included in the ablation study. The groups are specified
        either by being passed as a list of layer names, or by passing a `prefix` argument, denoting a prefix shared by
        all layer names in a group.
        :param prefix: A string that is a prefix of the names of a group that is already included in the ablation study.
        :type prefix: str
        :param args: Lists of strings (layer names) that correspond to groups that are already
        included in the ablation study.
        :type args: list
        """

        if prefix is not None:
            if type(prefix) is str:
                if frozenset([prefix]) in self.included_groups:
                    self.included_groups.remove(frozenset([prefix]))
            else:
                raise ValueError("`prefix` argument of layers.exclude_groups() should either be "
                                 "a `NoneType` or a `str`, but it received {0} (of type '{1}'."
                                 .format(str(prefix), type(prefix).__name__))

        for arg in args:
            if type(arg) is list and len(arg) > 1:
                if frozenset(arg) in self.included_groups:
                    self.included_groups.remove(frozenset(arg))
            else:
                raise ValueError("layers.exclude_groups() only accepts a prefix string, or "
                                 "lists (with more than one element) "
                                 "of strings, but it received {0} (of type '{1}')."
                                 .format(str(arg), type(arg).__name__))

    def print_all(self):
        """
        Prints all single layers that are included in the current ablation study configuration.
        """
        if len(self.included_layers) > 0:
            print("Included single layers are: \n")
            for layer in self.included_layers:
                print(layer)
        else:
            print("There are no single layers in this ablation study configuration.")

    def print_all_groups(self):
        """
        Prints all layer groups that are included in the current ablation study configuration.
        """
        if len(self.included_groups) > 0:
            print("Included layer groups are: \n")
            for layer_group in self.included_groups:
                if len(layer_group) > 1:
                    print("--- Layer group " + str(list(layer_group)))
                elif len(layer_group) == 1:
                    print('---- All layers prefixed "' + str(list(layer_group)[0]) + '"')
        else:
            print("There are no layer groups in this ablation study configuration.")



from maggy.optimizer.abstractoptimizer import AbstractOptimizer
from maggy.optimizer.randomsearch import RandomSearch
from maggy.optimizer.asha import Asha
from maggy.optimizer.singlerun import SingleRun


import pytest
import time
import random

from maggy import Searchspace

def test_searchspace_init():

    sp = Searchspace(argument_param=('DOUBLE', [1, 5]), param2=('integer', [3, 4]))

    exp_get = [1, 5]

    assert sp.get('argument_param') == exp_get
    assert sp.argument_param == exp_get # pylint: disable=no-member

def test_searchspace_add():

    sp = Searchspace(argument_param=('DOUBLE', [1, 5]))

    with pytest.raises(ValueError) as excinfo:
        sp.add('argument_param', ('DOUBLE', [1, 5]))
    assert "Hyperparameter name is reserved" in str(excinfo.value)

    with pytest.raises(ValueError) as excinfo:
        # add tuple with too many elements
        sp.add('param', ('DOUBLE', [1, 5], 'too many'))
    assert "Hyperparameter tuple has to be of length two" in str(excinfo.value)

    with pytest.raises(ValueError) as excinfo:
        # add unknown type
        sp.add('param', ('FLOAT', [1, 5]))
    assert "Hyperparameter type is not of type " in str(excinfo.value)

    with pytest.raises(ValueError) as excinfo:
        # add empty region list
        sp.add('param', ('DOUBLE', []))
    assert "Hyperparameter feasible region list" in str(excinfo.value)

    with pytest.raises(AssertionError) as excinfo:
        # add incompatible type and feasible region
        sp.add('param', ('DOUBLE', [1, 5, 5]))
        sp.add('param2', ('INTEGER', [1, 5, 5]))
    assert "For DOUBLE or " in str(excinfo.value)

    with pytest.raises(AssertionError) as excinfo:
        # lower bound higher than upper bound
        sp.add('param', ('DOUBLE', [5, 1]))
        sp.add('param2', ('INTEGER', [4, 1]))
    assert "Lower bound " in str(excinfo.value)

    with pytest.raises(ValueError) as excinfo:
        # Non integer boundaries for integer type parameter
        sp.add('param2', ('INTEGER', [1.5, 5]))
    assert "type INTEGER need to be integer:" in str(excinfo.value)

    with pytest.raises(ValueError) as excinfo:
        # Non numeric interval boundaries
        sp.add('param2', ('DOUBLE', ['lower', 5]))
    assert "type DOUBLE need to be integer or float:" in str(excinfo.value)




from maggy.core import config

if config.tf_version >= 2:
    from maggy.tb_experimental import write_hparams_proto

tensorboard_dir = None

def _register(trial_dir):
    global tensorboard_dir

    tensorboard_dir = trial_dir

def logdir():
    global tensorboard_dir
    return tensorboard_dir



import socket
import math
import numpy as np
from pyspark.sql import SparkSession
from pyspark import TaskContext

from hops import util as hopsutil
from hops import hdfs as hopshdfs


def _get_experiments_dir(name):
    """Checks if a directory for the given maggy experiment `name` exists and
    if not creates it and returns the path.
    """
    base_dir = hopshdfs._get_experiments_dir()
    maggy_exp_dir = base_dir + '/maggy/' + str(name)

    if not hopshdfs.exists(maggy_exp_dir):
        hopshdfs.mkdir(maggy_exp_dir)

    return maggy_exp_dir

def _get_run_dir(name):
    """Checks the index of the latest run of an experiments and creates a new
    directory with the index incremented by one.
    """
    maggy_exp_dir = _get_experiments_dir(name)

    ls_exp_dir = hopshdfs.ls(maggy_exp_dir)

    if len(ls_exp_dir) == 0:
        run_dir = maggy_exp_dir + '/run.1'
        hopshdfs.mkdir(run_dir)
        return 1, run_dir
    elif len(ls_exp_dir) > 0:
        run_index = list(map(lambda x: int(x.rsplit('.', 1)[-1]), ls_exp_dir))
        max_run = max(run_index)
        new_run = max_run + 1
        run_dir = maggy_exp_dir + '/run.' + str(new_run)
        hopshdfs.mkdir(run_dir)
        return new_run, run_dir

def _get_runs(name):
    """Returns a list of hdfs paths for the runs of the experiment with `name`.
    """
    maggy_exp_dir = _get_experiments_dir(name)
    return hopshdfs.ls(maggy_exp_dir)

def _init_run(run_dir, app_id):
    app_dir = run_dir + '/' + app_id
    trials = app_dir + '/trials'
    logs = app_dir + '/logs'

    hopshdfs.mkdir(trials)
    hopshdfs.mkdir(logs)

    return app_dir, trials, logs

def num_executors():
    """
    Get the number of executors configured for Jupyter
    Returns:
        Number of configured executors for Jupyter
    """
    sc = hopsutil._find_spark().sparkContext
    try:
        return int(sc._conf.get('spark.dynamicAllocation.maxExecutors'))
    except:
        return int(sc._conf.get('spark.executor.instances'))

def get_partition_attempt_id():
    """Returns partitionId and attemptNumber of the task context, when invoked
    on a spark executor.
    PartitionId is ID of the RDD partition that is computed by this task.
    The first task attempt will be assigned attemptNumber = 0, and subsequent
    attempts will have increasing attempt numbers.
    Returns:
        partitionId, attemptNumber -- [description]
    """
    task_context = TaskContext.get()
    return task_context.partitionId(), task_context.attemptNumber()

def print_trial_store(store):
    for _, value in store.items():
        print(value.to_json())

def _progress_bar(done, total):

            done_ratio = done/total
            progress = math.floor(done_ratio * 30)

            bar = '['

            for i in range(30):
                if i < progress:
                    bar += '='
                elif i == progress:
                    bar += '>'
                else:
                    bar += '.'

            bar += ']'
            return bar

def json_default_numpy(obj):
    if isinstance(obj, np.integer):
        return int(obj)
    elif isinstance(obj, np.floating):
        return float(obj)
    elif isinstance(obj, np.ndarray):
        return obj.tolist()
    else:
        raise TypeError("Object of type {0}: {1} is not JSON serializable"
            .format(type(obj), obj))


from maggy.ablation.ablator import AbstractAblator
from maggy.ablation.ablationstudy import AblationStudy
from maggy.core.exceptions import NotSupportedError
from hops import featurestore
from hops import hdfs as hopshdfs
import tensorflow as tf
from maggy.trial import Trial
import json


class LOCO(AbstractAblator):

    def __init__(self, ablation_study, final_store):
        super().__init__(ablation_study, final_store)
        self.base_dataset_generator = self.get_dataset_generator(ablated_feature=None)

    def get_number_of_trials(self):
        # plus one is for the base (reference) trial with all the components
        return len(self.ablation_study.features.included_features) + \
            len(self.ablation_study.model.layers.included_layers) +  \
            len(self.ablation_study.model.layers.included_groups) + 1

    def get_dataset_generator(self, ablated_feature=None, dataset_type='tfrecord', shuffle_buffer_size=10000):

        if self.ablation_study.custom_dataset_generator:
            pass
        else:
            training_dataset_name = self.ablation_study.hops_training_dataset_name
            training_dataset_version = self.ablation_study.hops_training_dataset_version
            label_name = self.ablation_study.label_name

            if dataset_type == 'tfrecord':

                def create_tf_dataset(num_epochs, batch_size):
                    dataset_dir = featurestore.get_training_dataset_path(training_dataset_name,
                                                                         training_dataset_version)
                    input_files = tf.gfile.Glob(dataset_dir + '/part-r-*')
                    dataset = tf.data.TFRecordDataset(input_files)
                    tf_record_schema = featurestore.get_training_dataset_tf_record_schema(training_dataset_name)
                    meta = featurestore.get_featurestore_metadata()
                    training_features = [feature.name
                                         for feature
                                         in meta.training_datasets[training_dataset_name +
                                                                   '_'
                                                                   + str(training_dataset_version)].features]

                    if ablated_feature is not None:
                        training_features.remove(ablated_feature)

                    training_features.remove(label_name)

                    def decode(example_proto):
                        example = tf.parse_single_example(example_proto, tf_record_schema)
                        # prepare the features
                        x = []
                        for feature_name in training_features:
                            # temporary fix for the case of tf.int types
                            if tf_record_schema[feature_name].dtype.is_integer:
                                x.append(tf.cast(example[feature_name], tf.float32))
                            else:
                                x.append(example[feature_name])

                        # prepare the labels
                        if tf_record_schema[label_name].dtype.is_integer:
                            y = [tf.cast(example[label_name], tf.float32)]
                        else:
                            y = [example[label_name]]

                        return x, y
                    dataset = dataset.map(decode).shuffle(shuffle_buffer_size).batch(batch_size).repeat(num_epochs)
                    return dataset

                return create_tf_dataset
            else:
                raise NotSupportedError(
                    "dataset type", dataset_type,
                    "Use 'tfrecord' or write your own custom dataset generator.")

    def get_model_generator(self, layer_identifier=None):

        base_model_generator = self.ablation_study.model.base_model_generator
        if layer_identifier is None:
            return base_model_generator

        def model_generator():
            base_model = base_model_generator()

            list_of_layers = [base_layer for base_layer in base_model.get_config()['layers']]
            if type(layer_identifier) is str:
                # ablation of a single layer
                for base_layer in reversed(list_of_layers[1:-1]):
                    # the first (input) and last (output) layers should not be considered, hence list_of_layers[1:-1]
                    if base_layer['config']['name'] == layer_identifier:
                        list_of_layers.remove(base_layer)
            elif type(layer_identifier) is set:
                # ablation of a layer group - all the layers in the group should be removed together
                if len(layer_identifier) > 1:
                    # group of layers (non-prefix)
                    for base_layer in reversed(list_of_layers[1:-1]):
                        if base_layer['config']['name'] in layer_identifier:
                            list_of_layers.remove(base_layer)
                elif len(layer_identifier) == 1:
                    # layer_identifier is a prefix
                    prefix = list(layer_identifier)[0].lower()
                    for base_layer in reversed(list_of_layers[1:-1]):
                        if base_layer['config']['name'].lower().startswith(prefix):
                            list_of_layers.remove(base_layer)

            base_json = base_model.to_json()
            new_dict = json.loads(base_json)
            new_dict['config']['layers'] = list_of_layers
            new_json = json.dumps(new_dict)
            new_model = tf.keras.models.model_from_json(new_json)

            return new_model
        return model_generator

    def initialize(self):
        """
        Prepares all the trials for LOCO policy (Leave One Component Out).
        In total `n+1` trials will be generated where `n` is equal to the number of components
        (e.g. features and layers) that are included in the ablation study
        (i.e. the components that will be removed one-at-a-time). The first trial will include all the components and
        can be regarded as the base for comparison.
        """

        # 0 - add first trial with all the components (base/reference trial)
        self.trial_buffer.append(Trial(self.create_trial_dict(None, None), trial_type='ablation'))

        # generate remaining trials based on the ablation study configuration:
        # 1 - generate feature ablation trials
        for feature in self.ablation_study.features.included_features:
            self.trial_buffer.append(Trial(self.create_trial_dict(ablated_feature=feature), trial_type='ablation'))

        # 2 - generate single-layer ablation trials
        for layer in self.ablation_study.model.layers.included_layers:
            self.trial_buffer.append(Trial(self.create_trial_dict(layer_identifier=layer), trial_type='ablation'))

        # 3 - generate layer-groups ablation trials
        # each element of `included_groups` is a frozenset of a set, so we cast again to get a set
        # why frozensets in the first place? because a set can only contain immutable elements
        # hence elements (layer group identifiers) are frozensets

        for layer_group in self.ablation_study.model.layers.included_groups:
            self.trial_buffer.append(
                Trial(self.create_trial_dict(layer_identifier=set(layer_group)), trial_type='ablation'))

    def get_trial(self, trial=None):
        if self.trial_buffer:
            return self.trial_buffer.pop()
        else:
            return None

    def finalize_experiment(self, trials):
        return

    def create_trial_dict(self, ablated_feature=None, layer_identifier=None):
        """
        Creates a trial dictionary that can be used for creating a Trial instance.

        :param ablated_feature: a string representing the name of a feature, or None
        :param layer_identifier: A string representing the name of a single layer, or a set representing a layer group.
        If the set has only one element, it is regarded as a prefix, so all layers with that prefix in their names
        would be regarded as a layer group. Otherwise, if the set has more than one element, the layers with
        names corresponding to those elements would be regarded as a layer group.
        :return: A trial dictionary that can be passed to maggy.Trial() constructor.
        :rtype: dict
        """

        trial_dict = {}
        if ablated_feature is None:
            trial_dict['dataset_function'] = self.base_dataset_generator
            trial_dict['ablated_feature'] = 'None'
        else:
            trial_dict['dataset_function'] = self.get_dataset_generator(ablated_feature, dataset_type='tfrecord')
            trial_dict['ablated_feature'] = ablated_feature

        if layer_identifier is None:
            trial_dict['model_function'] = self.ablation_study.model.base_model_generator
            trial_dict['ablated_layer'] = 'None'
        else:
            trial_dict['model_function'] = self.get_model_generator(layer_identifier=layer_identifier)
            if type(layer_identifier) is str:
                trial_dict['ablated_layer'] = layer_identifier
            elif type(layer_identifier) is set:
                if len(layer_identifier) > 1:
                    trial_dict['ablated_layer'] = str(list(layer_identifier))
                elif len(layer_identifier) == 1:
                    trial_dict['ablated_layer'] = 'Layers prefixed ' + str(list(layer_identifier)[0])

        return trial_dict



import statistics

from abc import ABC, abstractmethod


class AbstractEarlyStop(ABC):
    """An Abstract class to implement custom early stopping criteria."""

    @staticmethod
    @abstractmethod
    def earlystop_check(to_check, finalized_trials, direction):
        """A abstract static method that needs to be implemented with a custom
        early stopping criterium.

        The function is called internally in the user specified interval
        with three arguments. It is necessary to add these to the function
        definition.

        :param to_check: A dictionary of currently running
        trials, where the key is the `trial_id` and values are Trial objects.
        :type to_check: dictionary
        :param finalized_trials: A list of finalized Trial objects.
        :type finalized_trials: list
        :param direction: A string describing the search objective, i.e. 'min'
        or 'max'.
        :type direction: str
        """
        pass



from abc import ABC, abstractmethod


class AbstractAblator(ABC):

    def __init__(self, ablation_study, final_store):
        self.ablation_study = ablation_study
        self.final_store = final_store
        self.trial_buffer = []

    @abstractmethod
    def get_number_of_trials(self):
        """
        If applicable, calculate and return the total number of trials of the ablation experiment.
        Make sure to also include the base (reference) trial in the count.

        :return: total number of trials of the ablation study experiment
        :rtype: int
        """
        pass

    @abstractmethod
    def get_dataset_generator(self, ablated_feature, dataset_type='tfrecord'):
        """
        Create and return a dataset generator function based on the ablation policy to be used in a trial.
        The returned function will be executed on the executor per each trial.

        :param ablated_feature: the name of the feature to be excluded from the training dataset.
            Must match a feature name in the corresponding feature group in the feature store.
        :type ablated_feature: str
        :param dataset_type: type of the dataset. For now, we only support 'tfrecord'.
        :return: A function that generates a TFRecordDataset
        :rtype: function
        """
        pass

    @abstractmethod
    def get_model_generator(self, ablated_layer):
        pass

    @abstractmethod
    def initialize(self):
        """
        Initialize the ablation study experiment by generating a number of trials. Depending on the ablation policy,
        this method might generate all the trials (e.g. as in LOCO), or generate a number of trials to warm-start the
        experiment. The trials should be added to `trial_buffer` in form of `Trial` objects.
        """
        pass

    @abstractmethod
    def get_trial(self, ablation_trial=None):
        """
        Return a `Trial` to be assigned to an executor, or `None` if there are no trials remaining in the experiment.
        The trial should contain a dataset generator and a model generator.
        Depending on the ablator policy, the trials could come from a list (buffer) of pre-made trials,
        or generated on the fly.

        :rtype: `Trial` or `None`
        """
        pass

    @abstractmethod
    def finalize_experiment(self, trials):
        """
        This method will be called before finishing the experiment. Developers can implement this method
        e.g. for cleanup or extra logging.
        """
        pass




from abc import ABC, abstractmethod


class AbstractOptimizer(ABC):

    def __init__(self):
        self.searchspace = None
        self.num_trials = None
        self.final_store = None

    @abstractmethod
    def initialize(self):
        """
        A hook for the developer to initialize the optimizer.
        """
        pass

    @abstractmethod
    def get_suggestion(self, trial=None):
        """
        Return a `Trial` to be assigned to an executor, or `None` if there are
        no trials remaining in the experiment.

        :rtype: `Trial` or `None`
        """
        pass

    @abstractmethod
    def finalize_experiment(self, trials):
        """
        This method will be called before finishing the experiment. Developers
        can implement this method e.g. for cleanup or extra logging.
        """
        pass



""" pytest fixtures that can be resued across tests. the filename needs to be conftest.py
"""

# make sure env variables are set correctly
import findspark  # this needs to be the first import
findspark.init()

import logging
import pytest

from pyspark import HiveContext
from pyspark import SparkConf
from pyspark import SparkContext
from pyspark.streaming import StreamingContext


def quiet_py4j():
    """ turn down spark logging for the test context """
    logger = logging.getLogger('py4j')
    logger.setLevel(logging.WARN)


def pytest_addoption(parser):
    parser.addoption(
        "--spark-master", action="store", default=None, help='spark-master: "spark://name.local:7077"'
    )


@pytest.fixture(scope="session")
def sc(request):
    """ fixture for creating a spark context
    Args:
        request: pytest.FixtureRequest object
    """

    assert request.config.getoption("--spark-master") is not None, \
        'No Spark Master Address provided, use --spark-master: "spark://host:port" '

    conf = (SparkConf().setMaster(request.config.getoption("--spark-master"))
            .setAppName("pytest-pyspark-local-testing")
            .set("spark.dynamicAllocation.maxExecutors", 2)
            .set("spark.executor.instances", 2))
    scont = SparkContext(conf=conf)
    request.addfinalizer(lambda: scont.stop())

    quiet_py4j()
    return scont


@pytest.fixture(scope="session")
def hive_context(sc):
    """  fixture for creating a Hive Context. Creating a fixture enables it to be reused across all
        tests in a session
    Args:
        spark_context: spark_context fixture
    Returns:
        HiveContext for tests
    """
    return HiveContext(sc)


@pytest.fixture(scope="session")
def streaming_context(sc):
    return StreamingContext(sc, 1)


"""
API Module for the user to include in his training code.

"""
import threading
from datetime import datetime

from maggy.core import exceptions, config

from hops import hdfs as hopshdfs

class Reporter(object):
    """
    Thread-safe store for sending a metric and logs from executor to driver
    """

    def __init__(self, log_file, partition_id, task_attempt, print_executor):
        self.metric = None
        self.lock = threading.RLock()
        self.stop = False
        self.trial_id = None
        self.logs = ''
        self.log_file = log_file
        self.partition_id = partition_id
        self.task_attempt = task_attempt
        self.print_executor = print_executor

        #Open File desc for HDFS to log
        if not hopshdfs.exists(log_file):
            hopshdfs.dump('', log_file)
        self.fd = hopshdfs.open_file(log_file, flags='w')

    # report
    def broadcast(self, metric):
        """Broadcast a metric to the experiment driver with the heartbeat.

        :param metric: Metric to be broadcasted
        :type metric: int, float
        :raises exception: EarlyStopException if told by the experiment driver
        """
        with self.lock:
            # if stop == True -> raise exception to break training function
            self.metric = metric
            if self.stop:
                raise exceptions.EarlyStopException(metric)

    def log(self, log_msg, verbose=True):
        """Logs a message to the executor logfile and executor stderr and
        optionally prints the message in jupyter.

        :param log_msg: Message to log.
        :type log_msg: str
        :param verbose: Print in Jupyter Notebook, defaults to True
        :type verbose: bool, optional
        """
        with self.lock:
            msg = datetime.now().isoformat() + \
                ' (' + str(self.partition_id) + '/' + \
                str(self.task_attempt) + '): ' + str(log_msg)
            self.fd.write((msg + '\n').encode())
            jupyter_log = str(self.partition_id) + ': ' + log_msg
            if verbose:
                self.logs = self.logs + jupyter_log + '\n'
            else:
                self.print_executor(msg)

    def get_data(self):
        """Returns the metric and logs to be sent to the experiment driver.
        """
        with self.lock:
            log_to_send = self.logs
            self.logs = ''
            return self.metric, log_to_send

    def reset(self):
        """
        Resets the reporter to the initial state in order to start a new
        trial.
        """
        with self.lock:
            self.metric = None
            self.stop = False
            self.trial_id = None
            self.fd.flush()

    def early_stop(self):
        with self.lock:
            if self.metric is not None:
                self.stop = True

    def get_trial_id(self):
        with self.lock:
            return self.trial_id

    def set_trial_id(self, trial_id):
        with self.lock:
            self.trial_id = trial_id



from maggy.optimizer import AbstractOptimizer
from maggy.trial import Trial


class SingleRun(AbstractOptimizer):

    def __init__(self):
        super().__init__()
        self.trial_buffer = []

    def initialize(self):
        for _ in range(self.num_trials):
            self.trial_buffer.append(Trial({}))

    def get_suggestion(self, trial=None):
        if self.trial_buffer:
            return self.trial_buffer.pop()
        else:
            return None

    def finalize_experiment(self, trials):
        return



from maggy.optimizer import AbstractOptimizer
from maggy.searchspace import Searchspace
from maggy.trial import Trial


class RandomSearch(AbstractOptimizer):

    def __init__(self):
        super().__init__()
        self.trial_buffer = []

    def initialize(self):

        if (Searchspace.DOUBLE not in self.searchspace.names().values() and
                Searchspace.INTEGER not in self.searchspace.names().values()):
            raise NotImplementedError(
                "Searchspace needs at least one continuous parameter for random search.")

        list_of_random_trials = self.searchspace.get_random_parameter_values(self.num_trials)
        for parameters_dict in list_of_random_trials:
            self.trial_buffer.append(Trial(parameters_dict, trial_type='optimization'))

    def get_suggestion(self, trial=None):
        if self.trial_buffer:
            return self.trial_buffer.pop()
        else:
            return None

    def finalize_experiment(self, trials):
        return



import statistics
from maggy.earlystop import AbstractEarlyStop


class MedianStoppingRule(AbstractEarlyStop):
    """The Median Stopping Rule implements the simple strategy of stopping a
    trial if its performance falls below the median of other trials at similar
    points in time.
    """
    @staticmethod
    def earlystop_check(to_check, finalized_trials, direction):

        stop = []

        for trial_id, trial in to_check.items():

            results = []
            median = None

            # count step from zero so it can be used as index for array
            step = len(trial.metric_history)

            if step > 0:

                for fin_trial in finalized_trials:

                    if len(fin_trial.metric_history) >= step:
                        avg = sum(fin_trial.metric_history[:step])/float(step)
                        results.append(avg)

                try:
                    median = statistics.median(results)
                except statistics.StatisticsError as e:
                    raise Exception(
                        "Warning: StatisticsError when calling early stop method\n{}".format(e))

                if median is not None:
                    if direction == 'max':
                        if max(trial.metric_history) < median:
                            stop.append(trial_id)
                    elif direction == 'min':
                        if min(trial.metric_history) > median:
                            stop.append(trial_id)

        return stop




import tensorflow as tf
from tensorboard.plugins.hparams import api as hp

def create_hparams_config(sp):
    hparams = []

    for k, v in sp.names().items():
        if v == 'DOUBLE':
            hparams.append(hp.HParam(k, hp.RealInterval(float(sp.get(k)[0]), float(sp.get(k)[1]))))
        elif v == 'INTEGER':
            hparams.append(hp.HParam(k, hp.IntInterval(sp.get(k)[0], sp.get(k)[1])))
        elif v == 'DISCRETE':
            hparams.append(hp.HParam(k, hp.Discrete(sp.get(k))))
        elif v == 'CATEGORICAL':
            hparams.append(hp.HParam(k, hp.Discrete(sp.get(k))))

    return hparams

def write_hparams_proto(logdir, searchspace):

    HPARAMS = create_hparams_config(searchspace)
    METRICS = [
        hp.Metric(
            "epoch_accuracy",
            group="validation",
            display_name="accuracy (val.)",
        ),
        hp.Metric(
            "epoch_loss",
            group="validation",
            display_name="loss (val.)",
        ),
        hp.Metric(
            "batch_accuracy",
            group="train",
            display_name="accuracy (train)",
        ),
        hp.Metric(
            "batch_loss",
            group="train",
            display_name="loss (train)",
        ),
    ]

    with tf.summary.create_file_writer(logdir).as_default():
        hp.hparams_config(hparams=HPARAMS, metrics=METRICS)




class EarlyStopException(Exception):

    def __init__(self, metric):
        super().__init__()

        self.metric = metric


class NotSupportedError(Exception):
    """Raised when we are dealing with a situation that we do not (yet) support, e.g., a specific dataset type."""
    def __init__(self, category, value, suggestion=""):
        self.message = "({0}: {1}) is not supported by Maggy.{2}".format(category, value, suggestion)
        super().__init__(self.message)





from maggy.earlystop import AbstractEarlyStop


class NoStoppingRule(AbstractEarlyStop):
    """The no stopping rule never stops any trials early.
    """
    @staticmethod
    def earlystop_check(to_check, finalized_trials, direction):
        stop = []
        return stop



"""
The experiment driver implements the functionality for scheduling trials on maggy.
"""
import queue
import threading
import json
import os
import secrets
from datetime import datetime
from sys import maxsize as integer_max
from maggy import util
from maggy.optimizer import AbstractOptimizer, RandomSearch, Asha, SingleRun
from maggy.core import rpc
from maggy.trial import Trial
from maggy.earlystop import AbstractEarlyStop, MedianStoppingRule, NoStoppingRule
from maggy.searchspace import Searchspace

from maggy.ablation.ablator import AbstractAblator
from maggy.ablation.ablator.loco import LOCO
from maggy.ablation.ablationstudy import AblationStudy

from hops import constants as hopsconstants
from hops import hdfs as hopshdfs
from hops import util as hopsutil

driver_secret = None


class ExperimentDriver(object):

    SECRET_BYTES = 8

    # @Moritz:
    # for now, we infer the experiment type (an optimization experiment or an ablation study)
    # using keyword arguments, and set self.experiment_type to an according string.
    # Some of these arguments are required for any maggy experiment:
    # num_trials, name, num_executors, hb_interval, description, app_dir, log_dir, trial_dir
    # while some are specific to the type of experiment. For example, if the ExperimentDriver constructor
    # is called with a `searchspace` parameter, we infer that it is a hyperparameter optimization task,
    # and if it is called with `ablationstudy` parameter, we infer that it's an ablation study.

    def __init__(self, experiment_type, **kwargs):

        global driver_secret

        # COMMON EXPERIMENT SETUP
        self._final_store = []
        self._trial_store = {}
        self.num_executors = kwargs.get('num_executors')
        self._message_q = queue.Queue()
        self.name = kwargs.get('name')
        self.experiment_done = False
        self.worker_done = False
        self.hb_interval = kwargs.get('hb_interval')
        self.description = kwargs.get('description')
        self.experiment_type = experiment_type
        self.es_interval = kwargs.get('es_interval')
        self.es_min = kwargs.get('es_min')

        # TYPE-SPECIFIC EXPERIMENT SETUP
        if self.experiment_type == 'optimization':
            # set up an optimization experiment

            self.num_trials = kwargs.get('num_trials')

            searchspace = kwargs.get('searchspace')
            if isinstance(searchspace, Searchspace):
                self.searchspace = searchspace
            elif searchspace is None:
                self.searchspace = Searchspace()
            else:
                raise Exception(
                    "The experiment's search space should be an instance of maggy.Searchspace, "
                    "but it is {0} (of type '{1}')."
                    .format(str(searchspace), type(searchspace).__name__))

            optimizer = kwargs.get('optimizer')
            
            if optimizer is None:
                if len(self.searchspace.names()) == 0:
                    self.optimizer = SingleRun()
                else:
                    raise Exception(
                        'Searchspace has to be empty or None to use without optimizer')
            elif isinstance(optimizer, str):
                if optimizer.lower() == 'randomsearch':
                    self.optimizer = RandomSearch()
                elif optimizer.lower() == 'asha':
                    self.optimizer = Asha()
                elif optimizer.lower() == 'none':
                    if len(self.searchspace.names()) == 0:
                        self.optimizer = SingleRun()
                    else:
                        raise Exception(
                            "Searchspace has to be empty or None to use without Optimizer.")
                else:
                    raise Exception(
                        "Unknown Optimizer. Can't initialize experiment driver.")
            elif isinstance(optimizer, AbstractOptimizer):
                self.optimizer = optimizer
                print("Custom Optimizer initialized.")
            else:
                raise Exception(
                    "The experiment's optimizer should either be an string indicating the name "
                    "of an implemented optimizer (such as 'randomsearch') or an instance of "
                    "maggy.optimizer.AbstractOptimizer, "
                    "but it is {0} (of type '{1}')."
                    .format(str(optimizer), type(optimizer).__name__))
            # Set references to data in optimizer
            self.optimizer.num_trials = self.num_trials
            self.optimizer.searchspace = self.searchspace
            self.optimizer.final_store = self._final_store

            direction = kwargs.get('direction')
            if isinstance(direction, str) and direction.lower() in ['min', 'max']:
                self.direction = direction.lower()
            else:
                raise Exception(
                    "The experiment's direction should be an string (either 'min' or 'max') "
                    "but it is {0} (of type '{1}')."
                    .format(str(direction), type(direction).__name__))

            es_policy = kwargs.get('es_policy')
            if isinstance(es_policy, str):
                if es_policy.lower() == 'median':
                    self.earlystop_check = MedianStoppingRule.earlystop_check
                elif es_policy.lower() == 'none':
                    self.earlystop_check = NoStoppingRule.earlystop_check
                else:
                    raise Exception(
                        "The experiment's early stopping policy should either be a string ('median' or 'none') "
                        "or a custom policy that is an instance of maggy.earlystop.AbstractEarlyStop, "
                        "but it is {0} (of type '{1}')."
                        .format(str(es_policy), type(es_policy).__name__))
            elif isinstance(es_policy, AbstractEarlyStop):
                self.earlystop_check = es_policy.earlystop_check
                print("Custom Early Stopping policy initialized.")
            else:
                raise Exception(
                    "The experiment's early stopping policy should either be a string ('median' or 'none') "
                    "or a custom policy that is an instance of maggy.earlystop.AbstractEarlyStop, "
                    "but it is {0} (of type '{1}')."
                    .format(str(es_policy), type(es_policy).__name__))

            self.es_interval = kwargs.get('es_interval')
            self.es_min = kwargs.get('es_min')

            self.result = {'best_val': 'n.a.',
                           'num_trials': 0,
                           'early_stopped': 0}

        elif self.experiment_type == 'ablation':
            # set up an ablation study experiment
            self.earlystop_check = NoStoppingRule.earlystop_check

            ablation_study = kwargs.get('ablation_study')
            if isinstance(ablation_study, AblationStudy):
                self.ablation_study = ablation_study
            else:
                raise Exception(
                    "The experiment's ablation study configuration should be an instance of "
                    "maggy.ablation.AblationStudy, "
                    "but it is {0} (of type '{1}')."
                    .format(str(ablation_study), type(ablation_study).__name__))

            ablator = kwargs.get('ablator')
            if isinstance(ablator, str):
                if ablator.lower() == 'loco':
                    self.ablator = LOCO(ablation_study, self._final_store)
                    self.num_trials = self.ablator.get_number_of_trials()
                    if self.num_executors > self.num_trials:
                        self.num_executors = self.num_trials
                else:
                    raise Exception(
                        "The experiment's ablation study policy should either be a string ('loco') "
                        "or a custom policy that is an instance of maggy.ablation.ablation.AbstractAblator, "
                        "but it is {0} (of type '{1}')."
                        .format(str(ablator), type(ablator).__name__))
            elif isinstance(ablator, AbstractAblator):
                self.ablator = ablator
                print("Custom Ablator initialized. \n")
            else:
                raise Exception(
                    "The experiment's ablation study policy should either be a string ('loco') "
                    "or a custom policy that is an instance of maggy.ablation.ablation.AbstractAblator, "
                    "but it is {0} (of type '{1}')."
                    .format(str(ablator), type(ablator).__name__))

            self.result = {'best_val': 'n.a.',
                           'num_trials': 0,
                           'early_stopped': 'n.a'}
        else:
            raise Exception(
                "Unknown experiment type. experiment_type should be either 'optimization' or 'ablation', "
                "but it is {0}."
                .format(str(self.experiment_type)))

        # FINALIZE EXPERIMENT SETUP
        self.server = rpc.Server(self.num_executors)
        if not driver_secret:
            driver_secret = self._generate_secret(ExperimentDriver.SECRET_BYTES)
        self._secret = driver_secret
        self.job_start = datetime.now()
        self.executor_logs = ''
        self.maggy_log = ''
        self.log_lock = threading.RLock()
        self.log_file = kwargs.get('log_dir') + '/maggy.log'
        self.trial_dir = kwargs.get('trial_dir')
        self.app_dir = kwargs.get('app_dir')
        self.worker_exception = None

    # Open File desc for HDFS to log
        if not hopshdfs.exists(self.log_file):
            hopshdfs.dump('', self.log_file)
        self.fd = hopshdfs.open_file(self.log_file, flags='w')

    def init(self):

        self.server_addr = self.server.start(self)

        if self.experiment_type == 'optimization':
            self.optimizer.initialize()
        elif self.experiment_type == 'ablation':
            self.ablator.initialize()

        self._start_worker()

    def finalize(self, job_start, job_end):

        results = ''

        if self.experiment_type == 'optimization':

            _ = self.optimizer.finalize_experiment(self._final_store)

            self.job_end = datetime.now()

            self.duration = hopsutil._time_diff(self.job_start, self.job_end)

            results = '\n------ ' + str(self.optimizer.__class__.__name__) + ' Results ------ direction(' + self.direction + ') \n' \
                'BEST combination ' + json.dumps(self.result['best_hp']) + ' -- metric ' + str(self.result['best_val']) + '\n' \
                'WORST combination ' + json.dumps(self.result['worst_hp']) + ' -- metric ' + str(self.result['worst_val']) + '\n' \
                'AVERAGE metric -- ' + str(self.result['avg']) + '\n' \
                'EARLY STOPPED Trials -- ' + str(self.result['early_stopped']) + '\n' \
                'Total job time ' + self.duration + '\n'

        elif self.experiment_type == 'ablation':

            _ = self.ablator.finalize_experiment(self._final_store)
            self.job_end = datetime.now()
            self.duration = hopsutil._time_diff(self.job_start, self.job_end)

            results = "\n------ " + str(self.ablator.__class__.__name__) + " Results ------ \n" + \
                "BEST Config Excludes " + json.dumps(self.result['best_config']) + " -- metric " + \
                      str(self.result['best_val']) + "\n" + \
                "WORST Config Excludes " + json.dumps(self.result['worst_config']) + " -- metric " + \
                      str(self.result['worst_val']) + "\n" + \
                "AVERAGE metric -- " + str(self.result['avg']) + "\n" + \
                "Total Job Time " + self.duration + "\n"

        print(results)

        self._log(results)

        hopshdfs.dump(json.dumps(self.result, default=util.json_default_numpy),
                      self.app_dir + '/result.json')
        sc = hopsutil._find_spark().sparkContext
        hopshdfs.dump(self.json(sc), self.app_dir + '/maggy.json')

        return self.result

    def get_trial(self, trial_id):
        return self._trial_store[trial_id]

    def add_trial(self, trial):
        self._trial_store[trial.trial_id] = trial

    def add_message(self, msg):
        self._message_q.put(msg)

    def _start_worker(self):

        def _target_function(self):

            try:
                time_earlystop_check = datetime.now()  # only used by earlystop-supporting experiments
                while not self.worker_done:
                    trial = None
                    # get a message
                    try:
                        msg = self._message_q.get_nowait()
                    except:
                        msg = {'type': None}

                    if self.earlystop_check != NoStoppingRule.earlystop_check:
                        if (datetime.now() - time_earlystop_check).total_seconds() >= self.es_interval:
                            time_earlystop_check = datetime.now()

                        # pass currently running trials to early stop component
                            if len(self._final_store) > self.es_min:
                                self._log("Check for early stopping.")
                                try:
                                    to_stop = self.earlystop_check(
                                        self._trial_store, self._final_store, self.direction)
                                except Exception as e:
                                    self._log(e)
                                    to_stop = []
                                if len(to_stop) > 0:
                                    self._log("Trials to stop: {}".format(to_stop))
                                for trial_id in to_stop:
                                    self.get_trial(trial_id).set_early_stop()

                    # depending on message do the work
                    # 1. METRIC
                    if msg['type'] == 'METRIC':
                        # append executor logs if in the message
                        logs = msg.get('logs', None)
                        if logs is not None:
                            with self.log_lock:
                                self.executor_logs = self.executor_logs + logs

                        if msg['trial_id'] is not None and msg['data'] is not None:
                            self.get_trial(msg['trial_id']).append_metric(msg['data'])

                    # 2. BLACKLIST the trial
                    elif msg['type'] == 'BLACK':
                        trial = self.get_trial(msg['trial_id'])
                        with trial.lock:
                            trial.status = Trial.SCHEDULED
                            self.server.reservations.assign_trial(
                                msg['partition_id'], msg['trial_id'])

                    # 3. FINAL
                    elif msg['type'] == 'FINAL':
                        # set status
                        # get trial only once
                        trial = self.get_trial(msg['trial_id'])

                        logs = msg.get('logs', None)
                        if logs is not None:
                            with self.log_lock:
                                self.executor_logs = self.executor_logs + logs

                        # finalize the trial object
                        with trial.lock:
                            trial.status = Trial.FINALIZED
                            trial.final_metric = msg['data']
                            trial.duration = hopsutil._time_diff(
                                trial.start, datetime.now())

                        # move trial to the finalized ones
                        self._final_store.append(trial)
                        self._trial_store.pop(trial.trial_id)

                        # update result dictionary
                        self._update_result(trial)
                        # keep for later in case tqdm doesn't work
                        self.maggy_log = self._update_maggy_log()
                        self._log(self.maggy_log)

                        hopshdfs.dump(trial.to_json(), self.trial_dir + '/' + trial.trial_id + '/trial.json')

                        # assign new trial
                        if self.experiment_type == 'optimization':
                            trial = self.optimizer.get_suggestion(trial)
                        elif self.experiment_type == 'ablation':
                            trial = self.ablator.get_trial(trial)
                        if trial is None:
                            self.server.reservations.assign_trial(
                                msg['partition_id'], None)
                            self.experiment_done = True
                        else:
                            with trial.lock:
                                trial.start = datetime.now()
                                trial.status = Trial.SCHEDULED
                                self.server.reservations.assign_trial(
                                    msg['partition_id'], trial.trial_id)
                                self.add_trial(trial)

                    # 4. REG
                    elif msg['type'] == 'REG':
                        if self.experiment_type == 'optimization':
                            trial = self.optimizer.get_suggestion()
                        elif self.experiment_type == 'ablation':
                            trial = self.ablator.get_trial()
                        if trial is None:
                            self.experiment_done = True
                        else:
                            with trial.lock:
                                trial.start = datetime.now()
                                trial.status = Trial.SCHEDULED
                                self.server.reservations.assign_trial(
                                    msg['partition_id'], trial.trial_id)
                                self.add_trial(trial)
            except Exception as worker_exception:
                # Exception can't be propagated to parent thread
                # therefore log the exception and fail experiment
                self._log('Worker Exception')
                self._log(worker_exception)
                self.worker_exception = worker_exception
                self.server.stop()


        t = threading.Thread(target=_target_function, args=(self,))
        t.daemon = True
        t.start()

    def stop(self):
        """Stop the Driver's worker thread and server."""
        self.worker_done = True
        self.server.stop()
        self.fd.flush()
        self.fd.close()
        if self.worker_exception:
            raise Exception(
                "Worker: {}".format(self.worker_exception))

    def json(self, sc):
        """Get all relevant experiment information in JSON format.
        """
        user = None
        if hopsconstants.ENV_VARIABLES.HOPSWORKS_USER_ENV_VAR in os.environ:
            user = os.environ[hopsconstants.ENV_VARIABLES.HOPSWORKS_USER_ENV_VAR]

        experiment_json = {'project': hopshdfs.project_name(),
            'user': user,
            'name': self.name,
            'module': 'maggy',
            'app_id': str(sc.applicationId),
            'start': self.job_start.isoformat(),
            'memory_per_executor': str(sc._conf.get("spark.executor.memory")),
            'gpus_per_executor': str(sc._conf.get("spark.executor.gpus")),
            'executors': self.num_executors,
            'logdir': self.trial_dir,
            # 'versioned_resources': versioned_resources,
            'description': self.description,
            'experiment_type': self.experiment_type,
        }

        if self.experiment_type == 'optimization':
            experiment_json['hyperparameter_space'] = json.dumps(self.searchspace.to_dict())
            experiment_json['function'] = self.optimizer.__class__.__name__
        elif self.experiment_type == 'ablation':
            experiment_json['ablation_study'] = json.dumps(self.ablation_study.to_dict())
            experiment_json['ablator'] = self.ablator.__class__.__name__

        if self.experiment_done:
            experiment_json['status'] = "FINISHED"
            experiment_json['finished'] = self.job_end.isoformat()
            experiment_json['duration'] = self.duration
            if self.experiment_type == 'optimization':
                experiment_json['hyperparameter'] = json.dumps(self.result['best_hp'])
            experiment_json['metric'] = self.result['best_val']

        else:
            experiment_json['status'] = "RUNNING"

        return json.dumps(experiment_json, default=util.json_default_numpy)

    def _generate_secret(self, nbytes):
        """Generates a secret to be used by all clients during the experiment
        to authenticate their messages with the experiment driver.
        """
        return secrets.token_hex(nbytes=nbytes)

    def _update_result(self, trial):
        """Given a finalized trial updates the current result's best and
        worst trial.
        """

        metric = trial.final_metric
        param_string = trial.params
        trial_id = trial.trial_id

        if self.experiment_type == 'optimization':
            # First finalized trial
            if self.result.get('best_id', None) is None:
                self.result = {
                    'best_id': trial_id, 'best_val': metric,
                    'best_hp': param_string, 'worst_id': trial_id,
                    'worst_val': metric, 'worst_hp': param_string,
                    'avg': metric, 'metric_list': [metric], 'num_trials': 1,
                    'early_stopped': 0,
                }

                if trial.early_stop:
                    self.result['early_stopped'] += 1

                return
            if self.direction == 'max':
                if metric > self.result['best_val']:
                    self.result['best_val'] = metric
                    self.result['best_id'] = trial_id
                    self.result['best_hp'] = param_string
                if metric < self.result['worst_val']:
                    self.result['worst_val'] = metric
                    self.result['worst_id'] = trial_id
                    self.result['worst_hp'] = param_string
            elif self.direction == 'min':
                if metric < self.result['best_val']:
                    self.result['best_val'] = metric
                    self.result['best_id'] = trial_id
                    self.result['best_hp'] = param_string
                if metric > self.result['worst_val']:
                    self.result['worst_val'] = metric
                    self.result['worst_id'] = trial_id
                    self.result['worst_hp'] = param_string

        elif self.experiment_type == 'ablation':

            # pop function values and trial_type from parameters, since we don't need them
            param_string.pop('dataset_function', None)
            param_string.pop('model_function', None)
            # First finalized trial
            if self.result.get('best_id', None) is None:
                self.result = {
                    'best_id': trial_id, 'best_val': metric, 'best_config': param_string,
                    'worst_id': trial_id, 'worst_val': metric, 'worst_config': param_string,
                    'avg': metric, 'metric_list': [metric], 'num_trials': 1,
                    'early_stopped': 0,
                }
                return

            # for ablation we always consider 'max' as "the direction"
            if metric > self.result['best_val']:
                self.result['best_id'] = trial_id
                self.result['best_val'] = metric
                self.result['best_config'] = param_string
            elif metric < self.result['worst_val']:
                self.result['worst_id'] = trial_id
                self.result['worst_val'] = metric
                self.result['worst_config'] = param_string

        # update results and average regardless of experiment type
        self.result['metric_list'].append(metric)
        self.result['num_trials'] += 1
        self.result['avg'] = sum(self.result['metric_list'])/float(
            len(self.result['metric_list']))

        if trial.early_stop:
            self.result['early_stopped'] += 1

    def _update_maggy_log(self):
        """Creates the status of a maggy experiment with a progress bar.
        """
        finished = self.result['num_trials']
        log = ''

        if self.experiment_type == 'optimization':
            log = 'Maggy Optimization ' + str(finished) + '/' + str(self.num_trials) + \
                ' (' + str(self.result['early_stopped']) + ') ' + \
                util._progress_bar(finished, self.num_trials) + ' - BEST ' + \
                json.dumps(self.result['best_hp']) + ' - metric ' + \
                str(self.result['best_val'])

        elif self.experiment_type == 'ablation':
            log = "Maggy Ablation " + str(finished) + "/" + str(self.num_trials) + \
                util._progress_bar(finished, self.num_trials) + ' - BEST Excludes' + \
                json.dumps(self.result['best_config']) + ' - metric ' + \
                str(self.result['best_val'])

        return log

    def _get_logs(self):
        """Return current experiment status and executor logs to send them to
        spark magic.
        """
        with self.log_lock:
            temp = self.executor_logs
            # clear the executor logs since they are being sent
            self.executor_logs = ''
            return self.result, temp

    def _log(self, log_msg):
        """Logs a string to the maggy driver log file.
        """
        msg = datetime.now().isoformat() + ': ' + str(log_msg)
        self.fd.write((msg + '\n').encode())



import json
import random


class Searchspace(object):
    """Create an instance of `Searchspace` from keyword arguments.

    The keyword arguments specify name-values pairs for the hyperparameters,
    where values are tuples of the form (type, list). Type is a string with
    one of the following values:

        - DOUBLE
        - INTEGER
        - DISCRETE
        - CATEGORICAL

    And the list in the tuple specifies either two values only, the start
    and end point of of the feasible interval for DOUBLE and INTEGER,
    or the discrete possible values for the types DISCRETE and CATEGORICAL.

    Sample usage:

    >>> # Define Searchspace
    >>> from maggy import Searchspace
    >>> # The searchspace can be instantiated with parameters
    >>> sp = Searchspace(kernel=('INTEGER', [2, 8]), pool=('INTEGER', [2, 8]))
    >>> # Or additional parameters can be added one by one
    >>> sp.add('dropout', ('DOUBLE', [0.01, 0.99]))

    The `Searchspace` object can also be initialized from a python dictionary:

    >>> sp_dict = sp.to_dict()
    >>> sp_new = Searchspace(**sp_dict)

    The parameter names are added as attributes of `Searchspace` object,
    so they can be accessed directly with the dot notation
    `searchspace._name_`.
    """

    DOUBLE = "DOUBLE"
    INTEGER = "INTEGER"
    DISCRETE = "DISCRETE"
    CATEGORICAL = "CATEGORICAL"

    def __init__(self, **kwargs):
        self._hparam_types = {}
        for name, value in kwargs.items():
            self.add(name, value)

    def add(self, name, value):
        """Adds {name, value} pair to hyperparameters.

        :param name: Name of the hyperparameter
        :type name: str
        :param value: A tuple of the parameter type and its feasible region
        :type value: tuple
        :raises ValueError: Hyperparameter name is reserved
        :raises ValueError: Hyperparameter feasible region in wrong format
        """
        if getattr(self, name, None) is not None:
            raise ValueError(
                "Hyperparameter name is reserved: {}".format(name))

        if isinstance(value, tuple) or isinstance(value, list):

            if len(value) != 2:
                raise ValueError("Hyperparameter tuple has to be of length "
                                 "two and format (type, list): {0}, {1}".format(name, value))

            param_type = value[0].upper()
            param_values = value[1]

            if param_type in [Searchspace.DOUBLE,
                              Searchspace.INTEGER,
                              Searchspace.DISCRETE,
                              Searchspace.CATEGORICAL]:

                if len(param_values) == 0:
                    raise ValueError("Hyperparameter feasible region list "
                                     "cannot be empty: {0}, {1}".format(name, param_values))

                if param_type in [Searchspace.DOUBLE,
                                  Searchspace.INTEGER]:
                    assert len(param_values) == 2, ("For DOUBLE or INTEGER type parameters, list "
                                                    "can only contain upper and lower bounds: {0}, {1}".format(
                                                        name, param_values
                                                    ))

                    if param_type == Searchspace.DOUBLE:
                        if (type(param_values[0]) not in [int, float] or
                                type(param_values[1]) not in [int, float]):
                            raise ValueError("Hyperparameter boundaries for type DOUBLE need to be integer "
                                             "or float: {}".format(name))
                    elif param_type == Searchspace.INTEGER:
                        if (type(param_values[0]) != int or
                                type(param_values[1]) != int):
                            raise ValueError("Hyperparameter boundaries for type INTEGER need to be integer: "
                                             "{}".format(name))

                    assert param_values[0] < param_values[1], ("Lower bound {0} must be "
                                                               "less than upper bound {1}: {2}".format(
                                                                   param_values[0], param_values[1], name
                                                               ))

                self._hparam_types[name] = param_type
                setattr(self, name, value[1])
            else:
                raise ValueError("Hyperparameter type is not of type DOUBLE, "
                                 "INTEGER, DISCRETE or CATEGORICAL: {}".format(name))

        else:
            raise ValueError("Value is not an appropriate tuple: {}"
                             .format(name)
                             )

        print("Hyperparameter added: {}".format(name))

    def to_dict(self):
        """Return the hyperparameters as a Python dictionary.

        :return: A dictionary with hyperparameter names as keys. The values are
            the hyperparameter values.
        :rtype: dict
        """
        return {n: (self._hparam_types[n], getattr(self, n)) for n in self._hparam_types.keys()}

    def names(self):
        """Returns the dictionary with the names and types of all
        hyperparameters.

        :return: Dictionary of hyperparameter names, with types as value
        :rtype: dict
        """
        return self._hparam_types

    def get(self, name, default=None):
        """Returns the value of `name` if it exists, else `default`."""
        if name in self._hparam_types:
            return getattr(self, name)

        return default

    def get_random_parameter_values(self, num):
        """Generate random parameter dictionaries, e.g. to be used for initializing an optimizer.

        :param num: number of random parameter dictionaries to be generated.
        :type num: int
        :raises ValueError: `num` is not an int.
        :return: a list containing parameter dictionaries
        :rtype: list
        """
        return_list = []
        for _ in range(num):
            params = {}
            for name, value in self.names().items():
                feasible_region = self.get(name)
                if value == Searchspace.DOUBLE:
                    params[name] = random.uniform(feasible_region[0],
                                                  feasible_region[1])
                elif value == Searchspace.INTEGER:
                    params[name] = random.randint(feasible_region[0],
                                                  feasible_region[1])
                elif value == Searchspace.DISCRETE:
                    params[name] = random.choice(feasible_region)
                elif value == Searchspace.CATEGORICAL:
                    params[name] = random.choice(feasible_region)
            return_list.append(params)

        return return_list

    def __contains__(self, name):
        return name in self._hparam_types

    def __str__(self):
        return json.dumps(self.to_dict(), sort_keys=True)



"""
Experiment module used for running asynchronous optimization tasks.

The programming model is that you wrap the code containing the model
training inside a wrapper function.
Inside that wrapper function provide all imports and parts that make up your
experiment, see examples below. Whenever a function to run an experiment is
invoked it is also registered in the Experiments service along with the
provided information.
"""
import socket
import os
import json
import atexit
import time
from datetime import datetime

from maggy import util, tensorboard
from maggy.core import rpc, trialexecutor, ExperimentDriver
from maggy.trial import Trial

from hops import util as hopsutil
from hops import hdfs as hopshdfs

app_id = None
running = False
run_id = None
elastic_id = 1
experiment_json = None


def lagom(map_fun, name,
          experiment_type='optimization', hb_interval=1,
          num_trials=None, searchspace=None, optimizer=None, direction=None,
          ablation_study=None, ablator=None,
          es_policy='median', es_interval=300, es_min=10, description=''):
    """Launches a maggy experiment, which depending on `experiment_type` can
    either be a hyperparameter optimization or an ablation study experiment.
    Given a search space, objective and a model training procedure `map_fun`
    (black-box function), an experiment is the whole process of finding the
    best hyperparameter combination in the search space, optimizing the
    black-box function. Currently maggy supports random search and a median
    stopping rule.

    **lagom** is a Swedish word meaning "just the right amount".

    :param map_fun: User defined experiment containing the model training.
    :type map_fun: function
    :param experiment_type: Type of Maggy experiment, either 'optimization' (default) or 'ablation'.
    :type experiment_type: str
    :param searchspace: A maggy Searchspace object from which samples are drawn.
    :type searchspace: Searchspace
    :param optimizer: The optimizer is the part generating new trials.
    :type optimizer: str, AbstractOptimizer
    :param direction: If set to ‘max’ the highest value returned will
        correspond to the best solution, if set to ‘min’ the opposite is true.
    :type direction: str
    :param num_trials: the number of trials to evaluate given the search space,
        each containing a different hyperparameter combination
    :type num_trials: int
    :param name: A user defined experiment identifier.
    :type name: str
    :param hb_interval: The heartbeat interval in seconds from trial executor
        to experiment driver, defaults to 1
    :type hb_interval: int, optional
    :param es_policy: The earlystopping policy, defaults to 'median'
    :type es_policy: str, optional
    :param es_interval: Frequency interval in seconds to check currently
        running trials for early stopping, defaults to 300
    :type es_interval: int, optional
    :param es_min: Minimum number of trials finalized before checking for
        early stopping, defaults to 10
    :type es_min: int, optional
    :param description: A longer description of the experiment.
    :type description: str, optional
    :raises RuntimeError: An experiment is currently running.
    :return: A dictionary indicating the best trial and best hyperparameter
        combination with it's performance metric
    :rtype: dict
    """
    global running
    if running:
        raise RuntimeError("An experiment is currently running.")

    try:
        global app_id
        global experiment_json
        global elastic_id
        global run_id
        running = True

        sc = hopsutil._find_spark().sparkContext
        app_id = str(sc.applicationId)
        app_dir = ''

        # Create the root dir if not existing
        _ = util._get_experiments_dir(name)
        # get run_id and run_dir
        run_id, run_dir = util._get_run_dir(name)
        # set elastic id to run_id
        elastic_id = run_id
        # trial dir will be for tensorboard
        app_dir, trial_dir, log_dir = util._init_run(run_dir, app_id)
        tensorboard._register(trial_dir)
        #tensorboard.write_hparams_proto(trial_dir, searchspace)
        #hopshdfs.dump('writing proto buf worked', log_dir+'/maggy.log')

        num_executors = util.num_executors()

        # start experiment driver
        if experiment_type == 'optimization':

            assert num_trials > 0, "number of trials should be greater than zero"
            if num_executors > num_trials:
                num_executors = num_trials

            exp_driver = ExperimentDriver(
                'optimization', searchspace=searchspace, optimizer=optimizer,
                direction=direction, num_trials=num_trials, name=name,
                num_executors=num_executors, hb_interval=hb_interval, es_policy=es_policy,
                es_interval=es_interval, es_min=es_min, description=description,
                app_dir=app_dir, log_dir=log_dir, trial_dir=trial_dir)

        elif experiment_type == 'ablation':
            exp_driver = ExperimentDriver(
                'ablation', ablation_study=ablation_study, ablator=ablator,
                name=name, num_executors=num_executors,
                hb_interval=hb_interval, description=description,
                app_dir=app_dir, log_dir=log_dir, trial_dir=trial_dir)
            # using exp_driver.num_executor since
            # it has been set using ablator.get_number_of_trials() in experiment.py
            if num_executors > exp_driver.num_executors:
                num_executors = exp_driver.num_executors
        else:
            running = False
            raise RuntimeError(
                "Unknown experiment_type:"
                "should be either 'optimization' or 'ablation', "
                "But it is '{0}'".format(str(experiment_type)))

        nodeRDD = sc.parallelize(range(num_executors), num_executors)

        # Make SparkUI intuitive by grouping jobs
        sc.setJobGroup("Maggy Experiment", "{}".format(name))
        exp_driver._log("Started Maggy Experiment: {0}, run {1}".format(name, run_id))

        exp_driver.init()

        server_addr = exp_driver.server_addr

        experiment_json = exp_driver.json(sc)
        hopsutil._put_elastic(hopshdfs.project_name(), app_id, run_id,
            experiment_json)

        # Force execution on executor, since GPU is located on executor
        job_start = datetime.now()
        nodeRDD.foreachPartition(trialexecutor._prepare_func(app_id, run_id, experiment_type,
                                 map_fun, server_addr, hb_interval, exp_driver._secret, app_dir))
        job_end = datetime.now()

        result = exp_driver.finalize(job_start, job_end)

        experiment_json = exp_driver.json(sc)
        hopsutil._put_elastic(hopshdfs.project_name(), app_id, elastic_id,
            experiment_json)

        exp_driver._log("Finished Experiment")

    except:
        _exception_handler()
        raise
    finally:
        # grace period to send last logs to sparkmagic
        # sparkmagic hb poll intervall is 5 seconds, therefore wait 6 seconds
        time.sleep(6)
        # cleanup spark jobs
        exp_driver.stop()
        elastic_id += 1
        running = False
        sc.setJobGroup("", "")

    return result


def _exception_handler():
    """

    Returns:

    """
    global running
    global experiment_json
    if running and experiment_json is not None:
        experiment_json = json.loads(experiment_json)
        experiment_json['status'] = "FAILED"
        experiment_json['finished'] = datetime.now().isoformat()
        experiment_json = json.dumps(experiment_json)
        hopsutil._put_elastic(hopshdfs.project_name(), app_id, elastic_id, experiment_json)

def _exit_handler():
    """

    Returns:

    """
    global running
    global experiment_json
    if running and experiment_json != None:
        experiment_json = json.loads(experiment_json)
        experiment_json['status'] = "KILLED"
        experiment_json['finished'] = datetime.now().isoformat()
        experiment_json = json.dumps(experiment_json)
        hopsutil._put_elastic(hopshdfs.project_name(), app_id, elastic_id, experiment_json)

atexit.register(_exit_handler)



import json
import threading
import hashlib

from maggy import util


class Trial(object):
    """A Trial object contains all relevant information about the evaluation
    of an hyperparameter combination.

    It is used as shared memory between
    the worker thread and rpc server thread. The server thread performs only
    lookups on the `early_stop` and `params` attributes.
    """

    PENDING = "PENDING"
    SCHEDULED = "SCHEDULED"
    RUNNING = "RUNNING"
    ERROR = "ERROR"
    FINALIZED = "FINALIZED"

    def __init__(self, params, trial_type='optimization'):
        """Create a new trial object from a hyperparameter combination
        ``params``.

        :param params: A dictionary of Hyperparameters as key value pairs.
        :type params: dict
        """
        # XXX before merge, we should remove the default value for trial_type
        # and make sure everywhere Trial() is called (e.g. in all optimizers)
        # trial_type is passed
        # @Moritz

        self.trial_type = trial_type
        # XXX temp fix, have to come up with abstractions
        if self.trial_type == 'optimization':
            self.trial_id = Trial._generate_id(params)
        elif self.trial_type == 'ablation':
            serializable_params = {'ablated_feature': params.get('ablated_feature', None),
                                   'ablated_layer': params.get('ablated_layer', None),
                                   }
            self.trial_id = Trial._generate_id(serializable_params)
        self.params = params
        self.status = Trial.PENDING
        self.early_stop = False
        self.final_metric = None
        self.metric_history = []
        self.start = None
        self.duration = None
        self.lock = threading.RLock()

    def get_early_stop(self):
        """Return the early stopping flag of the trial."""
        with self.lock:
            return self.early_stop

    def set_early_stop(self):
        """Set the early stopping flag of the trial to true."""
        with self.lock:
            self.early_stop = True

    def append_metric(self, metric):
        """Append a metric from the heartbeats to the history."""
        with self.lock:
            self.metric_history.append(metric)

    @classmethod
    def _generate_id(cls, params):
        """
        Class method to generate a hash from a hyperparameter dictionary.

        All keys in the dictionary have to be strings. The hash is a to 16
        characters truncated md5 hash and stable across processes.

        :param params: Hyperparameters
        :type params: dictionary
        :raises ValueError: All hyperparameter names have to be strings.
        :raises ValueError: Hyperparameters need to be a dictionary.
        :return: Sixteen character truncated md5 hash
        :rtype: str
        """

        # ensure params is a dictionary
        if isinstance(params, dict):
            # check that all keys are strings
            if False in set(isinstance(k, str) for k in params.keys()):
                raise ValueError(
                    'All hyperparameter names have to be strings.')

            return hashlib.md5(
                json.dumps(params, sort_keys=True).encode('utf-8')
                ).hexdigest()[:16]

        raise ValueError("Hyperparameters need to be a dictionary.")

    def to_json(self):
        return json.dumps(self.to_dict(), default=util.json_default_numpy)

    def to_dict(self):
        obj_dict = {
            "__class__": self.__class__.__name__
        }

        temp_dict = self.__dict__.copy()
        temp_dict.pop('lock')
        temp_dict.pop('start')

        obj_dict.update(temp_dict)

        return obj_dict

    @classmethod
    def from_json(cls, json_str):
        """Creates a Trial instance from a previously json serialized Trial
        object instance.

        :param json_str: String containing the object.
        :type json_str: str
        :raises ValueError: json_str is not a Trial object.
        :return: Instantiated object instance of Trial.
        :rtype: Trial
        """

        temp_dict = json.loads(json_str)
        if temp_dict.get('__class__', None) != 'Trial':
            raise ValueError("json_str is not a Trial object.")
        if temp_dict.get('params', None) is not None:
            instance = cls(temp_dict.get('params'))
            instance.trial_id = temp_dict['trial_id']
            instance.status = temp_dict['status']
            instance.early_stop = temp_dict.get('early_stop', False)
            instance.final_metric = temp_dict['final_metric']
            instance.metric_history = temp_dict['metric_history']
            instance.duration = temp_dict['duration']

        return instance



import pytest
import time
import random

from maggy.searchspace import Searchspace
from maggy.optimizer import RandomSearch
from maggy import experiment

# this allows using the fixture in all tests in this module
pytestmark = pytest.mark.usefixtures("sc")

def test_randomsearch_init():

    sp = Searchspace(argument_param=('DOUBLE', [1, 5]), param2=('integer', [3, 4]))

    rs = RandomSearch(5, sp, [])

    assert rs.num_trials == 5
    assert rs.searchspace == sp

def test_randomsearch_initialize():

    sp = Searchspace(argument_param=('DOUBLE', [1, 5]), param2=('integer', [3, 4]))

    rs = RandomSearch(5, sp, [])

    rs.initialize()

    assert len(rs.trial_buffer) == 5

def test_rs_initialize2():

    sp = Searchspace(argument_param=('DISCRETE', [1, 5]))

    rs = RandomSearch(5, sp, [])

    with pytest.raises(NotImplementedError) as excinfo:
        rs.initialize()
    assert "Searchspace needs at least one continuous parameter" in str(excinfo.value)

def test_randomsearch(sc):

    sp = Searchspace(argument_param=('DOUBLE', [1, 5]))

    def train(argument_param, reporter):

        print("Entered train function with param {}".format(argument_param))

        for i in range(5):
            acc = i+random.random()
            reporter.broadcast(metric=acc)
            reporter.log("Metric: {}".format(acc))

            time.sleep(argument_param)

        return acc

    result = experiment.lagom(train, sp, optimizer='randomsearch', direction='max', num_trials=25, name='test', hb_interval=1, es_interval=10)
    assert type(result) == type({})



import os
import tensorflow as tf

HOPSWORKS = "HOPSWORKS"
SPARK_ONLY = "SPARK_ONLY"

mode = None
tf_full = tf.__version__.split(".")[0]
# for building the docs since mock object doesn't mock int()
if not isinstance(tf_full, str):
    tf_version = 2
else:
    tf_version = int(tf_full)

try:
    mode = os.environ['HOPSWORKS_VERSION']
    mode = HOPSWORKS
    print("You are running maggy on Hopsworks.")
except KeyError:
    mode = SPARK_ONLY
    print("You are running maggy in pure Spark mode.")


import threading
import struct
from pyspark import cloudpickle
import time
import select
import socket
import secrets
import json

from maggy.trial import Trial

from hops import constants as hopsconstants
from hops import util as hopsutil

MAX_RETRIES = 3
BUFSIZE = 1024 * 2

server_host_port = None


class Reservations(object):
    """Thread-safe store for worker reservations.

    Needs to be thread-safe mainly because the server listener thread can add
    reservations while the experiment driver might modify something on a
    reservation.
    """

    def __init__(self, required):
        """

        Args:
            required:
        """
        self.required = required
        self.lock = threading.RLock()
        self.reservations = {}
        self.check_done = False

    def add(self, meta):
        """
        Add a reservation.

        Args:
            :meta: a dictonary of metadata about a node
        """
        with self.lock:
            self.reservations[meta['partition_id']] = {
                'host_port': meta["host_port"],
                'task_attempt': meta['task_attempt'],
                'trial_id': meta['trial_id']}

            if self.remaining() == 0:
                self.check_done = True

    def done(self):
        """Returns True if the ``required`` number of reservations have been fulfilled."""
        with self.lock:
            return self.check_done

    def get(self):
        """Get the current reservations."""
        with self.lock:
            return self.reservations

    def remaining(self):
        """Get a count of remaining/unfulfilled reservations."""
        with self.lock:
            num_registered = len(self.reservations)
            return self.required - num_registered

    def get_assigned_trial(self, partition_id):
        """Get the ``trial_id`` of the trial assigned to ``partition_id``.

        Returns None if executor with ``partition_id`` is not registered or if
        ``partition_id`` is not assigned a trial yet.

        Args:
            :partition_id: An id to identify the spark executor.

        Returns:
            trial_id
        """
        with self.lock:
            reservation = self.reservations.get(partition_id, None)
            if reservation is not None:
                return reservation.get('trial_id', None)

    def assign_trial(self, partition_id, trial_id):
        """Assigns trial with ``trial_id`` to the reservation with ``partition_id``.

        Args:
            partition_id --
            trial {[type]} -- [description]
        """
        with self.lock:
            self.reservations.get(partition_id, None)['trial_id'] = trial_id


class MessageSocket(object):
    """Abstract class w/ length-prefixed socket send/receive functions."""

    def receive(self, sock):
        """
        Receive a message on ``sock``

        Args:
            sock:

        Returns:

        """
        msg = None
        data = b''
        recv_done = False
        recv_len = -1
        while not recv_done:
            buf = sock.recv(BUFSIZE)
            if buf is None or len(buf) == 0:
                raise Exception("socket closed")
            if recv_len == -1:
                recv_len = struct.unpack('>I', buf[:4])[0]
                data += buf[4:]
                recv_len -= len(data)
            else:
                data += buf
                recv_len -= len(buf)
            recv_done = (recv_len == 0)

        msg = cloudpickle.loads(data)
        return msg

    def send(self, sock, msg):
        """
        Send ``msg`` to destination ``sock``.

        Args:
            sock:
            msg:

        Returns:

        """
        data = cloudpickle.dumps(msg)
        buf = struct.pack('>I', len(data)) + data
        sock.sendall(buf)


class Server(MessageSocket):
    """Simple socket server with length prefixed pickle messages"""
    reservations = None
    done = False

    def __init__(self, count):
        """

        Args:
            count:
        """
        assert count > 0
        self.reservations = Reservations(count)

    def await_reservations(self, sc, status={}, timeout=600):
        """
        Block until all reservations are received.

        Args:
            sc:
            status:
            timeout:

        Returns:

        """
        timespent = 0
        while not self.reservations.done():
            print("waiting for {} reservations."
                  .format(self.reservations.remaining()))
            # check status flags for any errors
            if 'error' in status:
                sc.cancelAllJobs()
            time.sleep(1)
            timespent += 1
            if (timespent > timeout):
                raise Exception(
                    "Timed out waiting for reservations to complete")
        print("All reservations completed")
        return self.reservations.get()

    def _handle_message(self, sock, msg, exp_driver):
        """
        Handles a  message dictionary. Expects a 'type' and 'data' attribute in
        the message dictionary.

        Args:
            sock:
            msg:

        Returns:

        """
        msg_type = msg['type']

        # Prepare message
        send = {}

        if msg_type == 'REG':
            # check if executor was registered before and retrieve lost trial
            lost_trial = self.reservations.get_assigned_trial(msg['partition_id'])
            if lost_trial is not None:
                # the trial or executor must have failed
                exp_driver.get_trial(lost_trial).status = Trial.ERROR
                # add a blacklist message to the worker queue
                fail_msg = {'partition_id': msg['partition_id'],
                    'type': 'BLACK',
                    'trial_id': lost_trial}
                self.reservations.add(msg['data'])
                exp_driver.add_message(fail_msg)
            else:
                # else add regular registration msg to queue
                self.reservations.add(msg['data'])
                exp_driver.add_message(msg)

            send['type'] = 'OK'
        elif msg_type == 'QUERY':
            send['type'] = 'QUERY'
            send['data'] = self.reservations.done()
        elif msg_type == 'METRIC':
            # add metric msg to the exp driver queue
            exp_driver.add_message(msg)

            if msg['trial_id'] is None:
                send['type'] = 'OK'
                MessageSocket.send(self, sock, send)
                return
            elif msg['trial_id'] is not None:
                if msg.get('data', None) is None:
                    send['type'] = 'OK'
                    MessageSocket.send(self, sock, send)
                    return

            # lookup executor reservation to find assigned trial
            trialId = msg['trial_id']
            # get early stopping flag for hyperparameter optimization trials
            flag = False
            if exp_driver.experiment_type == 'optimization':
                flag = exp_driver.get_trial(trialId).get_early_stop()

            if flag:
                send['type'] = 'STOP'
            else:
                send['type'] = 'OK'
        elif msg_type == 'FINAL':
            # reset the reservation to avoid sending the same trial again
            self.reservations.assign_trial(msg['partition_id'], None)

            send['type'] = 'OK'

            # add metric msg to the exp driver queue
            exp_driver.add_message(msg)
        elif msg_type == 'GET':
            # lookup reservation to find assigned trial
            trial_id = self.reservations.get_assigned_trial(msg['partition_id'])

            # trial_id needs to be none because experiment_done can be true but
            # the assigned trial might not be finalized yet
            if exp_driver.experiment_done and trial_id is None:
                send['type'] = "GSTOP"
            else:
                send['type'] = "TRIAL"

            send['trial_id'] = trial_id

            # retrieve trial information
            if trial_id is not None:
                send['data'] = exp_driver.get_trial(trial_id).params
                exp_driver.get_trial(trial_id).status = Trial.RUNNING
            else:
                send['data'] = None
        elif msg_type == 'LOG':
            # get data from experiment driver
            result, log = exp_driver._get_logs()

            send['type'] = "OK"
            if log:
                send['ex_logs'] = log
            else:
                send['ex_logs'] = None
            send['num_trials'] = exp_driver.num_trials
            send['to_date'] = result['num_trials']
            send['stopped'] = result['early_stopped']
            send['metric'] = result['best_val']
        else:
            send['type'] = "ERR"

        MessageSocket.send(self, sock, send)

    def get_assigned_trial_id(self, partition_id):
        """Returns the id of the assigned trial, given a ``partition_id``.

        Arguments:
            partition_id {[type]} -- [description]

        Returns:
            trial_id
        """
        return self.reservations.get_assigned_trial(partition_id)

    def start(self, exp_driver):
        """
        Start listener in a background thread.

        Returns:
            address of the Server as a tuple of (host, port)
        """
        global server_host_port

        server_sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        server_sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        if not server_host_port:
            server_sock.bind(('', 0))
            # hostname may not be resolvable but IP address probably will be
            host = hopsutil._get_ip_address()
            port = server_sock.getsockname()[1]
            server_host_port = (host, port)

            # register this driver with Hopsworks
            sc = hopsutil._find_spark().sparkContext
            app_id = str(sc.applicationId)

            method = hopsconstants.HTTP_CONFIG.HTTP_POST
            resource_url = hopsconstants.DELIMITERS.SLASH_DELIMITER + \
                        hopsconstants.REST_CONFIG.HOPSWORKS_REST_RESOURCE + hopsconstants.DELIMITERS.SLASH_DELIMITER + \
                        "maggy" + hopsconstants.DELIMITERS.SLASH_DELIMITER + "drivers"
            json_contents = {"hostIp": host,
                            "port": port,
                            "appId": app_id,
                            "secret" : exp_driver._secret }
            json_embeddable = json.dumps(json_contents)
            headers = {hopsconstants.HTTP_CONFIG.HTTP_CONTENT_TYPE: hopsconstants.HTTP_CONFIG.HTTP_APPLICATION_JSON}

            try:
                response = hopsutil.send_request(method, resource_url, data=json_embeddable, headers=headers)

                if (response.status_code // 100) != 2:
                    print("No connection to Hopsworks for logging.")
                    exp_driver._log("No connection to Hopsworks for logging.")
            except Exception as e:
                print("Connection failed to Hopsworks. No logging.")
                exp_driver._log(e)
                exp_driver._log("Connection failed to Hopsworks. No logging.")
        else:
            server_sock.bind(server_host_port)
        server_sock.listen(10)

        def _listen(self, sock, driver):
            CONNECTIONS = []
            CONNECTIONS.append(sock)

            while not self.done:
                read_socks, _, _ = select.select(
                    CONNECTIONS, [], [], 60)
                for sock in read_socks:
                    if sock == server_sock:
                        client_sock, client_addr = sock.accept()
                        CONNECTIONS.append(client_sock)
                        _ = client_addr
                    else:
                        try:
                            msg = self.receive(sock)

                            # raise exception if secret does not match
                            # so client socket gets closed
                            if not secrets.compare_digest(msg['secret'],
                                                          exp_driver._secret):
                                exp_driver._log("SERVER secret: {}".format(exp_driver._secret))
                                exp_driver._log("ERROR: wrong secret {}".format(msg['secret']))
                                raise Exception

                            self._handle_message(sock, msg, driver)
                        except Exception as e:
                            _ = e
                            sock.close()
                            CONNECTIONS.remove(sock)

            server_sock.close()

        t = threading.Thread(target=_listen, args=(self, server_sock, exp_driver))
        t.daemon = True
        t.start()

        return server_host_port

    def stop(self):
        """
        Stop the server's socket listener.
        """
        self.done = True


class Client(MessageSocket):
    """Client to register and await node reservations.

    Args:
        :server_addr: a tuple of (host, port) pointing to the Server.
    """
    def __init__(self, server_addr, partition_id, task_attempt, hb_interval, secret):
        # socket for main thread
        self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.sock.connect(server_addr)
        # socket for heartbeat thread
        self.hb_sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.hb_sock.connect(server_addr)
        self.server_addr = server_addr
        self.done = False
        self.client_addr = (hopsutil._get_ip_address(), self.sock.getsockname()[1])
        self.partition_id = partition_id
        self.task_attempt = task_attempt
        self.hb_interval = hb_interval
        self._secret = secret

    def _request(self, req_sock, msg_type, msg_data=None, trial_id=None, logs=None):
        """Helper function to wrap msg w/ msg_type."""
        msg = {}
        msg['partition_id'] = self.partition_id
        msg['type'] = msg_type
        msg['secret'] = self._secret

        if msg_type == 'FINAL' or msg_type == 'METRIC':
            msg['trial_id'] = trial_id
            if logs == '':
                msg['logs'] = None
            else:
                msg['logs'] = logs

        #if msg_data or ((msg_data == True) or (msg_data == False)):
        #    msg['data'] = msg_data
        msg['data'] = msg_data

        done = False
        tries = 0
        while not done and tries < MAX_RETRIES:
            try:
                MessageSocket.send(self, req_sock, msg)
                done = True
            except socket.error as e:
                tries += 1
                if tries >= MAX_RETRIES:
                    raise
                print("Socket error: {}".format(e))
                req_sock.close()
                req_sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                req_sock.connect(self.server_addr)

        resp = MessageSocket.receive(self, req_sock)

        return resp

    def close(self):
        """Close the client's sockets."""
        self.sock.close()
        self.hb_sock.close()

    def register(self, registration):
        """
        Register ``registration`` with server.

        Args:
            registration:

        Returns:

        """
        resp = self._request(self.sock, 'REG', registration)
        return resp

    def await_reservations(self):
        done = False
        while not done:
            done = self._request(self.sock, 'QUERY').get('data', False)
            time.sleep(1)
        print("All executors registered: {}".format(done))
        return done

    def start_heartbeat(self, reporter):

        def _heartbeat(self, report):

            while not self.done:

                metric, logs = report.get_data()

                resp = self._request(self.hb_sock,
                                    'METRIC',
                                     metric,
                                     report.get_trial_id(),
                                     logs)
                _ = self._handle_message(resp, report)

                # sleep one second
                time.sleep(self.hb_interval)

        t = threading.Thread(target=_heartbeat, args=(self, reporter))
        t.daemon = True
        t.start()

        reporter.log("Started metric heartbeat", False)

    def get_suggestion(self):
        """Blocking call to get new parameter combination."""
        while not self.done:
            resp = self._request(self.sock, 'GET')
            trial_id, parameters = self._handle_message(resp) or (None, None)

            if trial_id is not None:
                break
            time.sleep(1)
        return trial_id, parameters

    def stop(self):
        """Stop the Clients's heartbeat thread."""
        self.done = True

    def _handle_message(self, msg, reporter=None):
        """
        Handles a  message dictionary. Expects a 'type' and 'data' attribute in
        the message dictionary.

        Args:
            sock:
            msg:

        Returns:

        """
        msg_type = msg['type']
        # if response is STOP command, early stop the training
        if msg_type == 'STOP':
            reporter.early_stop()
        elif msg_type == 'GSTOP':
            print("Stopping experiment")
            self.done = True
        elif msg_type == 'TRIAL':
            return msg['trial_id'], msg['data']
        elif msg_type == 'ERR':
            print("Stopping experiment")
            self.done = True

    def finalize_metric(self, metric, reporter):
        # make sure heartbeat thread can't send between sending final metric
        # and resetting the reporter
        with reporter.lock:
            _, logs = reporter.get_data()
            resp = self._request(self.sock, 'FINAL', metric, reporter.get_trial_id(), logs)
            reporter.reset()
        return resp



import math

from maggy.optimizer import AbstractOptimizer
from maggy.searchspace import Searchspace
from maggy.trial import Trial


class Asha(AbstractOptimizer):
    """Implements the Asynchronous Successiv Halving Algorithm - ASHA
    (https://arxiv.org/abs/1810.05934). ASHA needs three additional parameters:
    'reduction_factor', 'resource_min' and 'resource_max'. To set custom values
    for these, initialize the optimizer first and pass it as an argument to
    'experiment.lagom()'.

    Sample usage:

    >>> # Import Asha optimizer
    >>> from maggy.optimizer import Asha
    >>> # Instantiate the optimizer with custom arguments
    >>> asha = Asha(3, 1, 9)
    >>> experiment.lagom(..., optimizer=asha, ...)
    """

    def __init__(self, reduction_factor=2, resource_min=1, resource_max=4):
        super().__init__()

        if reduction_factor < 2 or not isinstance(reduction_factor, int):
            raise Exception(
                "Can't initialize ASHA optimizer. 'reduction_factor'" + \
                "has to be an integer equal to or larger than 2: {}"
                .format(reduction_factor))
        else:
                self.reduction_factor = reduction_factor

        if not isinstance(resource_min, int):
            raise Exception(
                "Can't initialize ASHA optimizer. 'resource_min'" + \
                "not of type INTEGER.")
        if not isinstance(resource_max, int):
            raise Exception(
                "Can't initialize ASHA optimizer. 'resource_max'" + \
                "not of type INTEGER.")
        if resource_min >= resource_max:
            raise Exception(
                "Can't initialize ASHA optimizer. 'resource_min' is larger" + \
                "than 'resource_max'.")

        self.resource_min = resource_min
        self.resource_max = resource_max

    def initialize(self):

        # maps rung index k to trials in that rung
        self.rungs = {0: []}
        # maps rung index k to trial ids of trials that were promoted
        self.promoted = {0: []}

        self.max_rung = int(math.floor(math.log(
            self.resource_max/self.resource_min, self.reduction_factor)))

        assert self.num_trials >= self.reduction_factor**(self.max_rung + 1)

    def get_suggestion(self, trial=None):

        if trial is not None:
            # stopping criterium: one trial in max rung
            if self.max_rung in self.rungs:
                # return None to signal end to experiment driver
                return None

            # for each rung
            for k in range(self.max_rung-1, -1, -1):
                # if rung doesn't exist yet go one lower
                if k not in self.rungs:
                    continue

                # get top_k
                rung_finished = len([x for x in self.rungs[k] if x.status == Trial.FINALIZED])

                if (rung_finished//self.reduction_factor) - len(self.promoted.get(k,[])) > 0:
                    candidates = self._top_k(k, (rung_finished//self.reduction_factor))
                else:
                    candidates = []

                # if there are no candidates, check one rung below
                if not candidates:
                    continue

                # select all that haven't been promoted yet
                promotable = [t for t in candidates if t.trial_id not in self.promoted.get(k,[])]

                nr_promotable = len(promotable)
                if nr_promotable >= 1:
                    new_rung = k + 1
                    # sorted in decending order, take highest -> index 0
                    old_trial = promotable[0]
                    # make copy of params to be able to change resource
                    params = old_trial.params.copy()
                    params['resource'] = self.resource_min * (self.reduction_factor**new_rung)
                    promote_trial = Trial(params)

                    # open new rung if not exists
                    if new_rung in self.rungs:
                        self.rungs[new_rung].append(promote_trial)
                    else:
                        self.rungs[new_rung] = [promote_trial]

                    # remember promoted trial
                    if k in self.promoted:
                        self.promoted[k].append(old_trial.trial_id)
                    else:
                        self.promoted[k] = [old_trial.trial_id]

                    return promote_trial

        # else return random configuration in base rung
        params = self.searchspace.get_random_parameter_values(1)[0]
        # set resource to minimum
        params['resource'] = self.resource_min
        to_return = Trial(params)
        # add to bottom rung
        self.rungs[0].append(to_return)
        return to_return

    def finalize_experiment(self, trials):
        return

    def _top_k(self, rung_k, number):
        """Find top-`number` trials in `rung_k`.
        """
        if number > 0:
            filtered = [x for x in self.rungs[rung_k] if x.status == Trial.FINALIZED]
            filtered.sort(key=lambda x: x.final_metric, reverse=True)
            # return top k trials if finalized
            return filtered[:number]
        else:
            return []



import pytest
import time
import random

from maggy import Trial

def test_trial_init():

    trial = Trial({'param1': 5, 'param2': 'ada'})

    exp = {'param1': 5, 'param2': 'ada'}

    assert trial.params == exp
    assert trial.status == Trial.PENDING
    assert trial.trial_id == '3d1cc9fdb1d4d001'

def test_trial_serialization():

    trial = Trial({'param1': 5, 'param2': 'ada'})

    exp = {'param1': 5, 'param2': 'ada'}

    json_str = trial.to_json()

    new_trial = Trial.from_json(json_str)

    assert isinstance(new_trial, Trial)
    assert new_trial.params == exp
    assert new_trial.status == Trial.PENDING
    assert new_trial.trial_id == '3d1cc9fdb1d4d001'



import pytest
from operator import add

# this allows using the fixture in all tests in this module
pytestmark = pytest.mark.usefixtures("sc")

# Can also use a decorator such as this to use specific fixtures in specific functions
# @pytest.mark.usefixtures("spark_context", "hive_context")

def do_word_counts(lines):
    """ count of words in an rdd of lines """

    counts = (lines.flatMap(lambda x: x.split())
                  .map(lambda x: (x, 1))
                  .reduceByKey(add)
             ) 
    results = {word: count for word, count in counts.collect()}
    return results

# start function with test_ so pytest can discover them
def test_do_word_counts(sc):
    """ test that a single event is parsed correctly
    Args:
        spark_context: test fixture SparkContext
        hive_context: test fixture HiveContext
    """

    test_input = [
        ' hello spark ',
        ' hello again spark spark'
    ]

    input_rdd = sc.parallelize(test_input, 1)
    results = do_word_counts(input_rdd)
    
    expected_results = {'hello':2, 'spark':3, 'again':1}  
    assert results == expected_results



import pytest
import time
import random
from maggy import experiment
from maggy.searchspace import Searchspace
from maggy.optimizer import RandomSearch
from operator import add

# this allows using the fixture in all tests in this module
pytestmark = pytest.mark.usefixtures("sc")

def test_nr_executors(sc):

    executor_instances = int(sc._conf.get("spark.executor.instances"))
    expected_number = 2
    assert executor_instances == expected_number

def test_random_search(sc):

    sp = Searchspace(argument_param=('DOUBLE', [1, 5]))

    rs = RandomSearch(5, sp, [])

    exp_result = {'argument_param': 'DOUBLE'}

    assert sp.names() == exp_result
    assert rs.num_trials == 5
    assert rs.searchspace == sp
